#! /usr/bin/python
# -*- coding: utf-8 -*-

import sys, fst, argparse, unicodedata
import regex as re

#
# command line arguments
#
arg_parser = argparse.ArgumentParser(prog="gp_encode.py", description="Encodes words as sequence of graphem sequences for further treatment with a g2p conversion.")

# mapping
arg_parser.add_argument('-M', '--mapping', metavar='FILE', type=file, help='Possible grapheme-phoneme mappings', required=True)

# mode
arg_parser.add_argument('-m', "--mode", help='Operation mode', choices=['train', 'scan', 'align'], default='scan')

# p field
arg_parser.add_argument('-p', "--pf", help='Phoneme field', choices=[1, 2], default=1, type=int)

# input
arg_parser.add_argument('input', metavar='IN', help='Words (or grapheme-phoneme pairs in training mode) to be encoded', type=argparse.FileType('r'), default=sys.stdin)

# output
arg_parser.add_argument('-o', '--output', metavar='FILE', nargs='?', type=argparse.FileType('w'), default=sys.stdout, help='Output')

args = arg_parser.parse_args()

split_re = re.compile("(\X)", re.U)

def my_linear_chain(g,syms):
    t = fst.Transducer(syms,syms)
    src = 0
    for c in split_re.findall(g):
        dest = t.add_state()
        t.add_arc(src,dest,c,c)
        src = dest
    t[src].final = True
    return t

# generate possible segmentations for a string
def segment(g):
    global editor_syms, Ig

    t2 = my_linear_chain(g,syms=editor_syms)
    t3 = t2.compose(Ig)
    return t3

# generate possible transcriptions for a string
def expand(g):
    global editor_syms, Ip, Ip_r

    #splits = Ip_r.findall(g)
    #t3 = my_linear_chain(u"_".join(splits) + u"_",syms=editor_syms)
    t2 = my_linear_chain(g,syms=editor_syms)
    t3 = t2.compose(Ip)
    return t3

# align grapheme and phoneme sequence
def align(g,p):
	global gp_syms, gp, split_re

	t3 = segment(g)
        t3.project_output()

	if t3.num_arcs() == 0:
		sys.stderr.write("Empty segmentation: %s %s\n" % (g.encode("utf-8"), p.encode("utf-8")))
		return []

	t4 = expand(p)
        t4.project_output()

	if t4.num_arcs() == 0:
		sys.stderr.write("Empty expansion: %s %s\n" % (g.encode("utf-8"), p.encode("utf-8")))
		return []
		
	try:
		t5 = t3.compose(E)
                t6 = t5.compose(t4)
	except:
		print "Unknown symbol: %s %s" % (g.encode("utf-8"), p.encode("utf-8"))	
		sys.exit(1)
        #t6 = t6.determinize()
        #t6.minimize()
	try:
                in_segs = []
                out_segs = []
                paths = list(enumerate(t6.paths()))
                if len(paths) > 1:
		    #sys.stderr.write("Multiple alignments: %s %s\n" % (g.encode("utf-8"), p.encode("utf-8")))
                    pass
		for i, path in paths:
			cur_in  = u""
			cur_out = u""
			for arc in path:
				isym = t5.isyms.find(arc.ilabel)
				osym = t5.osyms.find(arc.olabel)
				if isym == u'|':
					in_segs.append(cur_in)
					cur_in = u""
				elif isym != u'ε':
					cur_in += isym
                                if osym == u'§':
                                        out_segs.append(cur_out)
                                        cur_out = u"" 
				elif osym !=  u'ε':
					cur_out += osym
			break
	except:
		#print "%s\t%s\t%s" % (g.encode("utf-8"),g.encode("utf-8"),p.encode("utf-8"))
		#for c in splits:
		#	print c, "|".join(pg_map[c])
		sys.stderr.write("Empty composition: %s %s\n" % (g.encode("utf-8"), p.encode("utf-8")))
		pass
        if len(in_segs) != len(out_segs):
		sys.stderr.write("Alignment error: %s %s\n" % (g.encode("utf-8"), p.encode("utf-8")))
		pass

	return [in_segs,out_segs]

#
# main
#

#
# parse mapping
#
editor_syms = fst.SymbolTable ()
E = fst.Transducer(editor_syms,editor_syms)
Ig = fst.Transducer(editor_syms,editor_syms)
Ip = fst.Transducer(editor_syms,editor_syms)

# build the transducer
regex = []
g_set = {}
p_set = {}

E.add_arc(1, 0, u'ε', u'§')
for line in args.mapping:
	line = line.decode("utf-8").strip()
        if not line or line.startswith(u"# "):
            continue
	fields = line.split("\t")
        g = split_re.findall(fields[0])
        p = split_re.findall(fields[1])

        # unknown graph sequence
        if not g_set.has_key(fields[0]):
            dst_E = E.add_state()
	    E.add_arc(0, dst_E, g[0], u'ε')
            dst_I = Ig.add_state()
	    Ig.add_arc(0, dst_I, g[0], g[0])
	    for c in g[1:]:
                src_E = dst_E
                dst_E = E.add_state()
		E.add_arc(src_E, dst_E, c, u'ε')
                src_I = dst_I
                dst_I = Ig.add_state()
		Ig.add_arc(src_I, dst_I, c, c)
            g_set[fields[0]] = dst_E
	    Ig.add_arc(dst_I, 0, u'ε', u'|')
        if not p_set.has_key(fields[1]):
            src_E = E.add_state()
            src_I = 0
            p_set[fields[1]] = src_E
	    for c in p[:-1]:
                dst_E = E.add_state()
		E.add_arc(src_E, dst_E, u'ε', c)
                src_E = dst_E
                dst_I = Ip.add_state()
		Ip.add_arc(src_I, dst_I, c, c)
                src_I = dst_I
	    E.add_arc(src_E, 1, u'ε', p[-1])
            dst_I = Ip.add_state()
	    Ip.add_arc(src_I, dst_I, p[-1], p[-1])
	    Ip.add_arc(dst_I, 0, u'ε', u'§')
	E.add_arc(g_set[fields[0]], p_set[fields[1]], u'|', u'ε')
E[0].final = True
Ig[0].final = True
Ip[0].final = True

Ip_r = re.compile(u"%s" % u"|".join(u"%s" % re.escape(pat) for pat in sorted(list(p_set), key=len, reverse=True)), re.UNICODE)
clean_re = re.compile(u"[·#+\u0326\\\\]",re.U)

#
# encode
#
for line in args.input:
	line = line.strip ()
	if not line:
            continue
        fields = line.decode("utf-8").split(u"\t")

        if args.mode == "scan":

            t3 = segment(fields[0])

            try:
                    for i, path in enumerate(t3.paths()):
                            #args.output.write("# %s, sgmt. %d\n" %(fields[0].encode("utf-8"),i))
                            cur_in  = u""
                            for arc in path:
                                    osym = t3.osyms.find(arc.olabel)
                                    if osym == u'|':
                                        args.output.write("%s\n" % cur_in.encode("utf-8"))
                                        cur_in = u""
                                    else:
                                        cur_in += osym
                            args.output.write("\n")
            except:
                    pass
        else:
            #aligned = align(clean_re.sub(u"",fields[0]),clean_re.sub(u"",fields[args.pf]))
            aligned = align(fields[0],fields[args.pf])
            if not aligned:
                continue
            if args.mode == "train":
                args.output.write("\n".join( "%s\t%s" % (aligned[0][i].encode("utf-8"),aligned[1][i].encode("utf-8")) for i in range(0, len(aligned[0]))))
            else:
                args.output.write(" ".join( "%s,%s" % (aligned[0][i].encode("utf-8"),aligned[1][i].encode("utf-8")) for i in range(0, len(aligned[0]))))
            args.output.write('\n')
        args.output.write("\n")
