#include <ngram/ngram-output.h>

#include <fst/fst.h>
#include <fst/shortest-distance.h>
#include <fst/determinize.h>
#include <fst/rmepsilon.h>

#include <fst/script/print.h>
#include <fst/script/print-impl.h>

#include <iostream>
#include <fstream>

#include "gp_apply_model_cmdline.h"

static gp_apply_model_cmdline_args_info args; ///< command line arguments.

double handle_alignment(std::string alignment, fst::StdMutableFst *fst, ngram::NGramOutput* ngram)
{
  fst::StdVectorFst* chain = new fst::StdVectorFst;
  fst::StdVectorFst* result = new fst::StdVectorFst;
  fst::StdVectorFst* dresult = new fst::StdVectorFst;
  chain->SetInputSymbols(fst->InputSymbols());
  chain->SetOutputSymbols(fst->OutputSymbols());
  result->SetInputSymbols(fst->InputSymbols());
  result->SetOutputSymbols(fst->OutputSymbols());
  dresult->SetInputSymbols(fst->InputSymbols());
  dresult->SetOutputSymbols(fst->OutputSymbols());

  //fst::StdMutableFst *symbol_fst = chain->Copy();
  double logprob = 0.0;
  //int words = 0;
  //int oovs = 10000;
  //int words_skipped = 0;

  fst::StdVectorFst::StateId lastState = chain->AddState();
  chain->SetStart(lastState);
  fst::StdVectorFst::StateId newState;
  std::string symbol = "";
  //std::string oov_symbol = "";

  for(std::string::const_iterator sid = alignment.begin(); sid != alignment.end(); ++sid)
  {
    char c = *sid;
    switch(c)
    {
      case ' ':
        if(!symbol.empty())
        {
          // add arc for symbol
          newState = chain->AddState();
          chain->AddArc(lastState,fst::StdArc(chain->InputSymbols()->Find(symbol),chain->OutputSymbols()->Find(symbol),0,newState));
          lastState = newState;
          symbol = "";
        }
        break;
      case '\n':
        if(!symbol.empty())
        {
          // add fsm for symbol
          newState = chain->AddState();
          chain->SetFinal(newState,0);

          chain->AddArc(lastState,fst::StdArc(chain->InputSymbols()->Find(symbol),chain->OutputSymbols()->Find(symbol),0,newState));
          //fst::FstPrinter<fst::StdArc> print (*chain, chain->InputSymbols(), chain->OutputSymbols(), NULL, true, true);
          //print.Print(&std::cerr, "standard error");

          // query
          //ngram->ApplyNGramToFst(chain, symbol_fst, false, false, 0, 348, 0, &logprob, &words, &oovs, &words_skipped);
          ngram->FailLMCompose(*chain, result, ngram::kSpecialLabel);
          //fst::Compose(*chain,*fst,result);
          //fst::RmEpsilon(result);
          //fst::Determinize(*result,dresult);
          return fst::ShortestDistance(*result).Value();
        }

        break;
      default:
        symbol += c;
        break;
    }
  }
  if(!symbol.empty())
  {
    // add fsm for symbol
    newState = chain->AddState();
    chain->SetFinal(newState,0);

    chain->AddArc(lastState,fst::StdArc(chain->InputSymbols()->Find(symbol),chain->OutputSymbols()->Find(symbol),0,newState));
    //fst::FstPrinter<fst::StdArc> print (*chain, chain->InputSymbols(), chain->OutputSymbols(), NULL, true, true);
    //print.Print(&std::cerr, "standard error");

    // query
    ngram->FailLMCompose(*chain, result, ngram::kSpecialLabel);
    //fst::Compose(*chain,*fst,result);
    //fst::RmEpsilon(result);
    //fst::Determinize(*result,dresult);
    return fst::ShortestDistance(*result).Value();
  }
  return logprob;
}

int main(int argc, char **argv)
{
  if (gp_apply_model_cmdline_parser2 (argc, argv, &args, 1, 1, 0) != 0) // override, initialize, check_req
    return 1;
  if (gp_apply_model_cmdline_parser_required ( &args, argv[0]) != 0)
    return 1;

  fst::StdMutableFst *fst = fst::StdMutableFst::Read(args.model_arg, true);
  ngram::NGramOutput *ngram = new ngram::NGramOutput(fst);
  ngram->MakePhiMatcherLM(ngram::kSpecialLabel);
  char buffer[10000];

  // no unnamed arguments, use stdin
  if (args.inputs_num == 0)
  {
    while(std::cin.good())
    {
      std::cin.getline(buffer,10000);
      std::string line(buffer);
      if(line.length() > 1)
        std::cout << line << '\t' << handle_alignment(line.substr(line.rfind('\t')+1),fst,ngram);
      std::cout << '\n';
    }
  }
  else
  {
    for(unsigned int i = 0; i < args.inputs_num; ++i)
    {
      std::ifstream in(args.inputs[i]);
      while(in.good())
      {
        in.getline(buffer,10000);
        std::string line(buffer);
        if(line.length() > 1)
          std::cout << line << '\t' << handle_alignment(line.substr(line.rfind('\t')+1),fst,ngram);
        std::cout << '\n';
      }
    }
  }


  delete ngram;
  return 0;
}
